@@ -3,22 +3,26 @@ from configparser import ConfigParser
 import json
 import os
 from typing import Dict, Tuple
-import dask
+import sys
+
 import fsspec
 import numpy as np
 import traceback
+from gcsfs.core import GCSFileSystem
 
 from hub.client.hub_control import HubControlClient
+from hub.codec import Base as BaseCodec
+from hub.codec import from_name as codec_from_name
 from hub.collections.tensor.core import Tensor
 from hub.collections.client_manager import get_client
 from hub.log import logger
-from hub.exceptions import PermissionException
-from hub.utils import _flatten
+from hub.exceptions import (
+    PermissionException,
+    HubDatasetNotFoundException,
+    ModuleNotInstalledException,
+)
 
-try:
-    import torch
-except ImportError:
-    pass
+from hub.utils import _flatten
 
 
 class Transform:
@@ -31,21 +35,21 @@ class Transform:
     def meta(self):
         """
         Provides the metadata for all tensors including shapes, dtypes, dtags and chunksize for each array in the form
-        
+
         Returns
         -------
-        returns 
+        returns
             dict of tensor
-            
+
         Examples
         -------
         >>> def meta()
         >>>     return {
         >>>         ...
         >>>         "tesnor_name":{
-        >>>             "shape": (1,256,256), 
-        >>>             "dtype": "uint8", 
-        >>>             "chunksize": 100, 
+        >>>             "shape": (1,256,256),
+        >>>             "dtype": "uint8",
+        >>>             "chunksize": 100,
         >>>             "dtag": "segmentation"
         >>>         }
         >>>         ...
@@ -57,17 +61,17 @@ class Transform:
     def forward(input):
         """
         Takes a an element of a list or sample from dataset and returns sample of the dataset
-        
+
         Parameters
         -------
         input
-            an element of list or dict of arrays  
-            
+            an element of list or dict of arrays
+
         Returns
         -------
         dict
             dict of numpy arrays
-        
+
         Examples
         -------
         >>> def forward(input):
@@ -83,23 +87,23 @@ DatasetGenerator = Transform
 
 
 def _numpy_to_tuple(arr: np.ndarray):
-    """ Converts numpy array to tuple of numpy arrays 
-    """
+    """Converts numpy array to tuple of numpy arrays"""
     return [np.array([t]) for t in arr]
 
 
-def _numpy_saver(fs: fsspec.AbstractFileSystem, filepath: str, array: np.ndarray):
-    """ Saves a single numpy array into filepath given specific filesystem
-    """
+def _numpy_saver(
+    fs: fsspec.AbstractFileSystem, filepath: str, array: np.ndarray, codec: BaseCodec
+):
+    """Saves a single numpy array into filepath given specific filesystem"""
     with fs.open(filepath, "wb") as f:
-        np.save(f, array, allow_pickle=True)
+        f.write(codec.encode(array))
 
 
 def _numpy_saver_multi(
     fs: fsspec.AbstractFileSystem, filepath: str, arrays: np.ndarray, offset: int
 ):
     for i in range(len(arrays)):
-        _numpy_saver(fs, os.path.join(filepath, f"{offset+i}.npy"), arrays[i : i + 1])
+        _numpy_saver(fs, f"{filepath}/{offset+i}.npy", arrays[i : i + 1])
     return len(arrays)
 
 
@@ -110,14 +114,12 @@ def _preprocess_meta_before_save(meta: dict):
 
 
 def _dask_shape(input_shape: Tuple[int]):
-    """ Dask accept np.nan value in shape if the axis length is not known, our API uses -1 for that, this function converts -1 to np.nan
-    """
+    """Dask accept np.nan value in shape if the axis length is not known, our API uses -1 for that, this function converts -1 to np.nan"""
     return (np.nan,) + input_shape[1:] if input_shape[0] == -1 else input_shape
 
 
 def _dict_to_tuple(d: dict):
-    """ Converts dict of lists into (flattened list of values, list of keys)
-    """
+    """Converts dict of lists into (flattened list of values, list of keys)"""
     keys = sorted(d.keys())
     lens = {len(d[key]) for key in keys}
     assert len(lens) == 1
@@ -126,15 +128,14 @@ def _dict_to_tuple(d: dict):
 
 
 def _tuple_to_dict(t: tuple, keys: tuple):
-    """ Converts (flattened list of values, list of keys) into dict of lists
-    """
+    """Converts (flattened list of values, list of keys) into dict of lists"""
     cnt = len(keys)
     assert len(t) % cnt == 0
     return {key: [t[i] for i in range(j, len(t), cnt)] for j, key in enumerate(keys)}
 
 
 def _load_creds(creds):
-    """ Loads credentials from "{creds}" cfg file if such exists
+    """Loads credentials from "{creds}" cfg file if such exists
     if creds is dict, then dict will be returned, assuming all credential data is in dict
     """
     if creds is None:
@@ -148,8 +149,7 @@ def _load_creds(creds):
 
 
 def _connect(tag):
-    """ Connects to the backend and receive credentials
-    """
+    """Connects to the backend and receive credentials"""
 
     creds = HubControlClient().get_config()
     dataset = HubControlClient().get_dataset_path(tag)
@@ -167,9 +167,8 @@ def _connect(tag):
     return path, creds
 
 
-def _load_fs_and_path(path, creds=None, session_creds=True):
-    """ Given url(path) and creds returns filesystem required for accessing that file + url's filepath in that filesystem
-    """
+def _load_fs_and_path(path, creds=None, session_creds=True, google_cloud_project=""):
+    """Given url(path) and creds returns filesystem required for accessing that file + url's filepath in that filesystem"""
     if (
         path.startswith("./")
         or path.startswith("/")
@@ -178,7 +177,12 @@ def _load_fs_and_path(path, creds=None, session_creds=True):
     ):
         return fsspec.filesystem("file"), os.path.expanduser(path.replace("fs://", ""))
 
-    if session_creds and creds is None and not path.startswith("s3://"):
+    if (
+        session_creds
+        and creds is None
+        and not path.startswith("s3://")
+        and not path.startswith("gcs://")
+    ):
         path, creds = _connect(path)
 
     if path.startswith("s3://"):
@@ -201,27 +205,42 @@ def _load_fs_and_path(path, creds=None, session_creds=True):
         elif creds is not None:
             return (
                 fsspec.filesystem(
-                    "s3", key=creds.get("access_key"), secret=creds.get("secret_key"),
+                    "s3",
+                    key=creds.get("access_key"),
+                    secret=creds.get("secret_key"),
                 ),
                 path,
             )
         else:
             return fsspec.filesystem("s3"), path
+    elif path.startswith("gcs://"):
+        return (
+            GCSFileSystem(project=google_cloud_project, token=creds),
+            path[6:],
+        )
 
 
 class Dataset:
-    def __init__(self, tensors: Dict[str, Tensor]):
-        """ Creates dict given dict of tensors (name -> Tensor key value pairs)
-        """
+    def __init__(self, tensors: Dict[str, Tensor], metainfo=dict()):
+        """Creates dict given dict of tensors (name -> Tensor key value pairs)"""
         self._tensors = tensors
+        self._metainfo = metainfo
         shape = None
         for name, tensor in tensors.items():
             if shape is None or tensor.ndim > len(shape):
                 shape = tensor.shape
             self._len = tensor.count
+        self.verison = "0.x"
+        if "dask" not in sys.modules:
+            raise ModuleNotInstalledException("dask")
+        else:
+            import dask
+            import dask.array
+
+            global dask
 
     def __len__(self) -> int:
-        """ len of dataset (len of tensors across axis 0, yes, they all should be = to each other) 
+        """len of dataset (len of tensors across axis 0, yes, they all should be = to each other)
         Raises Exception if length is unknown
         """
         if self._len == -1:
@@ -230,36 +249,52 @@ class Dataset:
             )
         return self._len
 
+    @property
+    def license(self) -> str:
+        """Dataset license"""
+        return self._metainfo.get("license") if self._metainfo else None
+
+    @property
+    def description(self) -> str:
+        """Dataset description"""
+        return self._metainfo.get("description") if self._metainfo else None
+
+    @property
+    def citation(self) -> str:
+        """Dataset citation"""
+        return self._metainfo.get("citation") if self._metainfo else None
+
+    @property
+    def howtoload(self) -> str:
+        """Dataset howtoload"""
+        return self._metainfo.get("howtoload") if self._metainfo else None
+
     @property
     def count(self) -> int:
-        """ len of dataset (len of tensors across axis 0, yes, they all should be = to each other) 
+        """len of dataset (len of tensors across axis 0, yes, they all should be = to each other)
         Returns -1 if length is unknown
         """
         return self._len
 
     def __iter__(self):
-        """ Iterates over axis 0 return dict of Tensors
-        """
+        """Iterates over axis 0 return dict of Tensors"""
         for i in range(len(self)):
             yield {key: t._array[i] for key, t in self._tensors.items()}
 
     def keys(self):
-        """ Returns names of tensors
-        """
+        """Returns names of tensors"""
         yield from self._tensors.keys()
 
     def values(self):
-        """ Returns tensors
-        """
+        """Returns tensors"""
         yield from self._tensors.values()
 
     def items(self):
-        """ Returns tensors
-        """
+        """Returns tensors"""
         yield from self._tensors.items()
 
     def __getitem__(self, slices) -> "Dataset":
-        """ Returns a slice of dataset
+        """Returns a slice of dataset
         slices can be
             1) List of strs (slicing horizontally)
             2) List of slices or ints (slicing vertically)
@@ -374,13 +409,15 @@ class Dataset:
             for key in list(collected.keys()):
                 c = collected[key]
                 chunksize = self._tensors[key].chunksize
+                codec = codec_from_name(self._tensors[key].dcompress)
                 cnt = len(c) - len(c) % chunksize if not lasttime else len(c)
                 for i in range(0, cnt, chunksize):
                     tasks += [
                         dask.delayed(_numpy_saver)(
                             fs,
-                            os.path.join(path, key, f"{collected_offset[key] + i}.npy"),
+                            f"{path}/{key}/{collected_offset[key] + i}.npy",
                             c[i : i + chunksize],
+                            codec,
                         )
                     ]
                 collected_offset[key] += cnt
@@ -420,16 +457,16 @@ class Dataset:
                     collected[el] = _dask_concat([collected[el], arr])
                 c = collected[el]
                 chunksize_ = self._tensors[el].chunksize
+                codec = codec_from_name(self._tensors[el].dcompress)
                 if len(c) >= chunksize_ or lasttime:
                     jcnt = len(c) - len(c) % chunksize_ if not lasttime else len(c)
                     for j in range(0, jcnt, chunksize_):
                         tasks += [
                             dask.delayed(_numpy_saver)(
                                 fs,
-                                os.path.join(
-                                    path, el, f"{collected_offset[el] + j}.npy"
-                                ),
+                                f"{path}/{el}/{collected_offset[el] + j}.npy",
                                 collected[el][j : j + chunksize_],
+                                codec,
                             )
                         ]
                     collected_offset[el] += jcnt
@@ -443,7 +480,7 @@ class Dataset:
 
     @property
     def meta(self) -> dict:
-        """ Dict of meta's of each tensor
+        """Dict of meta's of each tensor
         meta of tensor contains all metadata for tensor storage
         """
         tensor_meta = {
@@ -454,8 +491,7 @@ class Dataset:
         return ds_meta
 
     def delete(self, tag, creds=None, session_creds=True) -> bool:
-        """ Deletes dataset given tag(filepath) and credentials (optional)
-        """
+        """Deletes dataset given tag(filepath) and credentials (optional)"""
         fs, path = _load_fs_and_path(tag, creds, session_creds=session_creds)
         fs: fsspec.AbstractFileSystem = fs
         if fs.exists(path):
@@ -464,27 +500,26 @@ class Dataset:
         return False
 
     def store(self, tag, creds=None, session_creds=True) -> "Dataset":
-        """ Stores dataset by tag(filepath) given credentials (can be omitted)
-        """
+        """Stores dataset by tag(filepath) given credentials (can be omitted)"""
         fs, path = _load_fs_and_path(tag, creds, session_creds=session_creds)
         fs: fsspec.AbstractFileSystem = fs
 
         if (
-            fs.exists(os.path.join(path))
-            and not fs.exists(os.path.join(path, "meta.json"))
-            # and not fs.exists(os.path.join(path, "HUB_DATASET"))
-            # and len(fs.ls(path, detail=False)) > 0
+            fs.exists(path)
+            and not fs.exists(f"{path}/meta.json")
+            and not fs.exists(f"{path}/HUB_DATASET")
+            and len(fs.ls(path, detail=False)) > 0
         ):
             raise Exception(f"This path {path} is not a dataset path, tag: {tag}")
         self.delete(tag, creds)
         fs.makedirs(path)
 
-        # with fs.open(os.path.join(path, "HUB_DATASET"), "w") as f:
-        #     f.write("Hello World")
+        with fs.open(f"{path}/HUB_DATASET", "w") as f:
+            f.write("Hello World")
 
-        tensor_paths = [os.path.join(path, t) for t in self._tensors]
+        tensor_paths = [f"{path}/{t}" for t in self._tensors]
         for tensor_path in tensor_paths:
-            fs.makedir(tensor_path)
+            fs.makedirs(tensor_path)
         tensor_meta = {
             name: _preprocess_meta_before_save(t._meta)
             for name, t in self._tensors.items()
@@ -502,25 +537,43 @@ class Dataset:
         for _, el in tensor_meta.items():
             el["shape"] = (count,) + tuple(el["shape"][1:])
         ds_meta = {"tensors": tensor_meta, "len": count}
-        with fs.open(os.path.join(path, "meta.json"), "w") as f:
+        ds_info = dict()
+        for key, value in self._metainfo.items():
+            ds_info[key] = value
+        ds_meta["metainfo"] = ds_info
+        with fs.open(f"{path}/meta.json", "w") as f:
             f.write(json.dumps(ds_meta, indent=2, sort_keys=True))
 
         return load(tag, creds)
 
-    def to_pytorch(self, transform=None):
+    def to_pytorch(self, transform=None, max_text_len=30):
         """
-            Transforms into pytorch dataset
-            
-            Parameters
-            ----------
-            transform: func
-                any transform that takes input a dictionary of a sample and returns transformed dictionary 
-        """
-        return TorchDataset(self, transform)
+        Transforms into pytorch dataset
 
-    def to_tensorflow(self):
+        Parameters
+        ----------
+        transform: func
+            any transform that takes input a dictionary of a sample and returns transformed dictionary
+        max_text_len: integer
+            the maximum length of text strings that would be stored. Strings longer than this would be snipped
         """
-            Transforms into tensorflow dataset
+        try:
+            import torch
+
+            global torch
+        except ImportError:
+            pass
+        return TorchDataset(self, transform, max_text_len)
+
+    def to_tensorflow(self, max_text_len=30):
+        """
+        Transforms into tensorflow dataset
+
+        Parameters
+        ----------
+        max_text_len: integer
+            the maximum length of text strings that would be stored. Strings longer than this would be snipped
+
         """
         try:
             import tensorflow as tf
@@ -536,37 +589,69 @@ class Dataset:
                         arrs
                     )
                     arrs = arrs.compute()
+                    for ind, arr in enumerate(arrs):
+                        if arr.dtype.type is np.str_:
+                            arr = [
+                                ([ord(x) for x in sample.tolist()[0:max_text_len]])
+                                for sample in arr
+                            ]
+                            arr = np.array(
+                                [
+                                    np.pad(
+                                        sample,
+                                        (0, max_text_len - len(sample)),
+                                        "constant",
+                                        constant_values=(32),
+                                    )
+                                    for sample in arr
+                                ]
+                            )
+                            arrs[ind] = arr
+
                     for i in range(step):
                         sample = {key: r[i] for key, r in zip(self[index].keys(), arrs)}
                         yield sample
 
         def tf_dtype(np_dtype):
             try:
+                if "U" in np_dtype:
+                    return tf.dtypes.as_dtype("string")
                 return tf.dtypes.as_dtype(np_dtype)
             except Exception as e:
+                logger.log(e)
                 return tf.variant
 
+        output_shapes = {}
+        output_types = {}
+        for key in self.keys():
+            output_types[key] = tf_dtype(self._tensors[key].dtype)
+            output_shapes[key] = self._tensors[key].shape[1:]
+
+            # if this is a string, we change the type to int, as it's going to become ascii. shape is also set to None
+            if output_types[key] == tf.dtypes.as_dtype("string"):
+                output_types[key] = tf.dtypes.as_dtype("int8")
+                output_shapes[key] = None
+
         # TODO use None for dimensions you don't know the length tf.TensorShape([None])
         # FIXME Dataset Generator is not very good with multiprocessing but its good for fast tensorflow support
         return tf.data.Dataset.from_generator(
             tf_gen,
-            output_types={
-                key: tf_dtype(self._tensors[key].dtype) for key in self.keys()
-            },
-            output_shapes={key: self._tensors[key].shape[1:] for key in self.keys()},
+            output_types=output_types,
+            # output_shapes=output_shapes,
         )
 
 
-def _numpy_load(fs: fsspec.AbstractFileSystem, filepath: str) -> np.ndarray:
-    """ Given filesystem and filepath, loads numpy array
-    """
+def _numpy_load(
+    fs: fsspec.AbstractFileSystem, filepath: str, codec: BaseCodec
+) -> np.ndarray:
+    """Given filesystem and filepath, loads numpy array"""
     # assert fs.exists(
     #    filepath
     # ), f"Dataset file {filepath} does not exists. Your dataset data is likely to be corrupted"
 
     try:
         with fs.open(filepath, "rb") as f:
-            return np.load(f, allow_pickle=True)
+            return codec.decode(f.read())
     except Exception as e:
         logger.error(traceback.format_exc() + str(e))
         raise Exception(
@@ -574,26 +659,59 @@ def _numpy_load(fs: fsspec.AbstractFileSystem, filepath: str) -> np.ndarray:
         )
 
 
+def get_text(input):
+    """Converts strings stored as ascii value tensors back into strings"""
+    if input.ndim == 1:
+        try:
+            text = "".join([chr(x) for x in input]).rstrip()
+            return text
+        except Exception as e:
+            logger.error(traceback.format_exc() + str(e))
+            raise Exception(
+                "get_text can only be called on a tensor of text or a batch of tensors of text"
+            )
+    elif input.ndim == 2:
+        try:
+            text = ["".join([chr(x) for x in sample]).rstrip() for sample in input]
+            return text
+        except Exception as e:
+            logger.error(traceback.format_exc() + str(e))
+            raise Exception(
+                "get_text can only be called on a tensor of text or a batch of tensors of text"
+            )
+    else:
+        raise Exception(
+            f"Got input of dimension {input.ndim} for get_text. Expected dimension of 1 or 2"
+        )
+
+
 def load(tag, creds=None, session_creds=True) -> Dataset:
-    """ Load a dataset from repository using given url and credentials (optional)
-    """
+    """Load a dataset from repository using given url and credentials (optional)"""
     fs, path = _load_fs_and_path(tag, creds, session_creds=session_creds)
     fs: fsspec.AbstractFileSystem = fs
-    path_2 = os.path.join(path, "meta.json")
+    path_2 = f"{path}/meta.json"
     if not fs.exists(path):
-        from hub.exceptions import DatasetNotFound
-
-        raise DatasetNotFound(tag)
+        raise HubDatasetNotFoundException(tag)
 
     with fs.open(path_2, "r") as f:
         ds_meta = json.loads(f.read())
 
     for name in ds_meta["tensors"]:
         assert fs.exists(
-            os.path.join(path, name)
+            f"{path}/{name}"
         ), f"Tensor {name} of {tag} dataset does not exist"
+
+    if "dask" not in sys.modules:
+        raise ModuleNotInstalledException("dask")
+    else:
+        import dask
+        import dask.array
+
+        global dask
+
     if ds_meta["len"] == 0:
         logger.warning("The dataset is empty (has 0 samples)")
+
         return Dataset(
             {
                 name: Tensor(
@@ -603,7 +721,8 @@ def load(tag, creds=None, session_creds=True) -> Dataset:
                     ),
                 )
                 for name, tmeta in ds_meta["tensors"].items()
-            }
+            },
+            metainfo=ds_meta.get("metainfo"),
         )
     len_ = ds_meta["len"]
 
@@ -620,7 +739,9 @@ def load(tag, creds=None, session_creds=True) -> Dataset:
                     [
                         dask.array.from_delayed(
                             dask.delayed(_numpy_load)(
-                                fs, os.path.join(path, name, f"{i}.npy")
+                                fs,
+                                f"{path}/{name}/{i}.npy",
+                                codec_from_name(tmeta.get("dcompress")),
                             ),
                             shape=(min(tmeta["chunksize"], len_ - i),)
                             + tuple(tmeta["shape"][1:]),
@@ -631,7 +752,8 @@ def load(tag, creds=None, session_creds=True) -> Dataset:
                 ),
             )
             for name, tmeta in ds_meta["tensors"].items()
-        }
+        },
+        metainfo=ds_meta.get("metainfo"),
     )
 
 
@@ -646,17 +768,20 @@ def _is_tensor_dynamic(tensor):
     return str(tensor.dtype) == "object" and _is_arraylike(arr.flatten()[0])
 
 
-#
-
-
 class TorchDataset:
-    def __init__(self, ds, transform=None):
+    def __init__(self, ds, transform=None, max_text_len=30):
         self._ds = ds
         self._transform = transform
         self._dynkeys = {
             key for key in self._ds.keys() if _is_tensor_dynamic(self._ds[key])
         }
-        self._client = None
+        self._max_text_len = max_text_len
+
+        def cost(nbytes, time):
+            print(nbytes, time)
+            return float(time) / (nbytes or 1) / 1e9
+
+        self.client = None
 
     def _do_transform(self, data):
         return self._transform(data) if self._transform else data
@@ -665,15 +790,19 @@ class TorchDataset:
         return len(self._ds)
 
     def __getitem__(self, index):
-        with dask.config.set(scheduler="sync"):
+        with dask.config.set(scheduler="sync", delayed_pure=True):
             arrs = [self._ds[index : index + 1].values() for i in range(1)]
             arrs = list(map(lambda x: x._array, _flatten(arrs)))
-            arrs = dask.delayed(list, pure=False, nout=len(list(self._ds.keys())))(arrs)
+            arrs = dask.delayed(list, pure=True, nout=len(list(self._ds.keys())))(arrs)
             arrs = arrs.compute()
+
             arrs = {key: r[0] for key, r in zip(self._ds[index].keys(), arrs)}
 
         objs = self._do_transform(arrs)
-        objs = {k: self._to_tensor(k, v) for k, v in objs.items()}
+        if isinstance(objs, dict):
+            objs = {k: self._to_tensor(k, v) for k, v in objs.items()}
+        elif isinstance(objs, list):
+            objs = [self._to_tensor(v) for v in objs]
         return objs
 
     def __iter__(self):
@@ -682,6 +811,16 @@ class TorchDataset:
 
     def _to_tensor(self, key, sample):
         if key not in self._dynkeys:
+            if isinstance(sample, np.str_):
+                sample = np.array(
+                    [ord(x) for x in sample.tolist()[0 : self._max_text_len]]
+                )
+                sample = np.pad(
+                    sample,
+                    (0, self._max_text_len - len(sample)),
+                    "constant",
+                    constant_values=(32),
+                )
             return torch.tensor(sample)
         else:
             return [torch.tensor(item) for item in sample]
@@ -703,33 +842,3 @@ def _dask_concat(arr):
         return arr[0]
     else:
         return dask.array.concatenate(arr)
-
-
-# class TensorflowDataset(tfds.core.GeneratorBasedBuilder):
-#     def _info(self):
-#         return tfds.core.DatasetInfo(
-#             builder=self,
-#             # This is the description that will appear on the datasets page.
-#             description=(
-#                 "This is the dataset for xxx. It contains yyy. The "
-#                 "images are kept at their original dimensions."
-#             ),
-#             # tfds.features.FeatureConnectors
-#             # features=tfds.features.FeaturesDict(
-#             #     {
-#             #         "image_description": tfds.features.Text(),
-#             #         "image": tfds.features.Image(),
-#             #         # Here, labels can be of 5 distinct values.
-#             #         "label": tfds.features.ClassLabel(num_classes=5),
-#             #     }
-#             # ),
-#             # If there's a common (input, target) tuple from the features,
-#             # specify them here. They'll be used if as_supervised=True in
-#             # builder.as_dataset.
-#             # supervised_keys=("image", "label"),
-#             # Homepage of the dataset for documentation
-#             homepage="https://dataset-homepage.org",
-#             # Bibtex citation for the dataset
-#             citation=r"""@article{my-awesome-dataset-2020,
-#                                 author = {Smith, John},"}""",
-#         )
