@@ -8,6 +8,7 @@ CACHE_FILE_PATH = os.path.expanduser("~/.activeloop/tmp")
 AWSCRED_PATH = os.path.expanduser("~/.aws/credentials")
 
 HUB_REST_ENDPOINT = "https://app.activeloop.ai"
+HUB_DEV_REST_ENDPOINT = "https://app.dev.activeloop.ai"
 HUB_LOCAL_REST_ENDPOINT = "http://localhost:5000"
 
 DEFAULT_TIMEOUT = 170
@@ -18,4 +19,7 @@ GET_REGISTER_SUFFIX = "/api/user/register"
 GET_DATASET_SUFFIX = "/api/dataset/get"
 GET_DATASET_PATH_SUFFIX = "/api/dataset/get/path"
 
-DISTRIBUTED = True
\ No newline at end of file
+CREATE_DATASET_SUFFIX = "/api/dataset/create"
+UPDATE_STATE_SUFFIX = "/api/dataset/state"
+DATASET_SUFFIX = "/api/dataset/"
+DISTRIBUTED = True
